
// Written Dec 2011 Tom Horsley
// Released to the universe with no license restrictions of any kind.

#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <vector>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include "packable.h"
#include "find-mkisofs.h"

std::vector<std::string> algos;

struct iso_formats {
   const char * name;
   unsigned long long max_sectors;
   const char * desc;
};

struct iso_formats sector_list[] = {
   {"DVD+R.SL",     2295104, "12 cm 1 Layer DVD+R"},
   {"DVD+R.DL",     4173824, "12 cm 2 Layer DVD+R"},
   {"DVD-R.SL",     2298496, "12 cm 1 Layer DVD-R"},
   {"DVD-R.DL",     4171712, "12 cm 2 Layer DVD-R"},
   {"BD-R.SL",     12219392, "12 cm 1 Layer Blu-Ray Recordable"},
   {"BD-R.DL",     24438784, "12 cm 2 Layer Blu-Ray Recordable"},
   {"BD-R.8cm.SL",  3804288, "8 cm 1 Layer Blu-Ray Recordable"},
   {"BD-R.8cm.DL",  7608576, "8 cm 2 Layer Blu-Ray Recordable"},
   {"BD-XL.TL",    48878592, "12 cm 3 Layer Blu-Ray XL Recordable"},
   {"BD-XL.QL",    62500864, "12 cm 4 Layer Blu-Ray XL Recordable"},
   {"DVD+RW",       2295104, "12 cm DVD+RW"},
   {"DVD-RW.F",     2297888, "12 cm DVD-RW formatted"},
   {"DVD-RW.U",     2298496, "12 cm DVD-RW not formatted"},
   {"BD-RE.F",     11826176, "12 cm 1 Layer Blu-Ray RE default format"},
   {"BD-RE.Z",     12219392, "12 cm 1 Layer Blu-Ray RE 0 spare format"},
};

#define NUM_ISO (sizeof(sector_list)/sizeof(sector_list[0]))

static bool read_packing_list_from(std::string infile);
static bool print_packing_to(std::string outfile);

static void algorithm_help();
static void time_help();
static void close_help();
static void waste_help();
static void watchdog_help();
static void verbose_help();
static void sectors_help();
static void prog_help();
static void args_help();
static void addarg_help();
static void infile_help();
static void outfile_help();
static void unpacked_help();
static void usage_help();

int
main(int argc, char ** argv) {
   std::string infile;
   std::string outfile;
   std::string unpacked;

   for (int i = 1; i < argc; ++i) {
      if (strncmp(argv[i], "-algorithm=", 11) == 0) {
         if (strcmp(argv[i] + 11, "help") == 0) {
            algorithm_help();
            return 0;
         }
         algos.push_back(argv[i] + 11);
      } else if (strncmp(argv[i], "-time=", 6) == 0) {
         if (strcmp(argv[i]+6, "help") == 0) {
            time_help();
            return 0;
         }
         char * endp = NULL;
         unsigned long cut = strtoul(argv[i] + 6, &endp, 10);
         if ((*endp == 'm') || (*endp == 'M')) {
            cut *= 60; // minutes * 60 => seconds
            Packable::SetSecondsLimit(cut);
         } else if ((*endp == 'h') || (*endp == 'H')) {
            cut *= (60*60); // hours * 60 => minutes * 60 => seconds
            Packable::SetSecondsLimit(cut);
         } else if (*endp == '\0') {
            Packable::SetSecondsLimit(cut);
         } else {
            std::cerr << "Unrecognized time format: " << argv[i] << std::endl;
            return 2;
         }
      } else if (strncmp(argv[i], "-close=", 7) == 0) {
         if (strcmp(argv[i]+7, "help") == 0) {
            close_help();
            return 0;
         }
         unsigned long long close_enough = 0;
         char * endp = NULL;
         close_enough = strtoull(argv[i]+7, &endp, 10);
         if (*endp != '\0') {
            std::cerr << "Not a number: " << argv[i] << std::endl;
            return 2;
         }
         Packable::SetCloseEnough(close_enough);
      } else if (strncmp(argv[i], "-waste=", 7) == 0) {
         if (strcmp(argv[i]+7, "help") == 0) {
            waste_help();
            return 0;
         }
         unsigned long long max_waste = 0;
         char * endp = NULL;
         max_waste = strtoull(argv[i]+7, &endp, 10);
         if (*endp != '\0') {
            std::cerr << "Not a number: " << argv[i] << std::endl;
            return 2;
         }
         Packable::SetMaxWaste(max_waste);
      } else if (strncmp(argv[i], "-watchdog=", 10) == 0) {
         if (strcmp(argv[i]+10, "help") == 0) {
            watchdog_help();
            return 0;
         }
         char * endp = NULL;
         unsigned long cut = strtoul(argv[i] + 10, &endp, 10);
         if ((*endp == 'm') || (*endp == 'M')) {
            cut *= 60; // minutes * 60 => seconds
            Packable::SetWatchdogSecondsLimit(cut);
         } else if ((*endp == 'h') || (*endp == 'H')) {
            cut *= (60*60); // hours * 60 => minutes * 60 => seconds
            Packable::SetWatchdogSecondsLimit(cut);
         } else if (*endp == '\0') {
            Packable::SetWatchdogSecondsLimit(cut);
         } else {
            std::cerr << "Unrecognized time format: " << argv[i] << std::endl;
            return 2;
         }
      } else if (strncmp(argv[i], "-verbose=", 9) == 0) {
         if (strcmp(argv[i] + 9, "help") == 0) {
            verbose_help();
            return 0;
         }
         unsigned long verb = strtoul(argv[i] + 9, NULL, 10);
         Packable::SetVerbose((int)verb);
      } else if (strncmp(argv[i], "-sectors=", 9) == 0) {
         if (strcmp(argv[i]+9, "help") == 0) {
            sectors_help();
            return 0;
         }
         unsigned long long sectors = 0;
         for (int j = 0; j < NUM_ISO; ++j) {
            if (strcmp(argv[i]+9, sector_list[j].name) == 0) {
               sectors = sector_list[j].max_sectors;
            }
         }
         if (sectors == 0) {
            char * endp = NULL;
            sectors = strtoull(argv[i]+9, &endp, 10);
            if (*endp != '\0') {
               std::cerr << "Unrecognized sector size: " << argv[i] << std::endl;
               return 2;
            }
         }
         Packable::SetMaxSectors(sectors);
      } else if (strncmp(argv[i], "-prog=", 6) == 0) {
         if (strcmp(argv[i]+6, "help") == 0) {
            prog_help();
            return 0;
         }
         FindMkisofs::SetMkisofsProgram(argv[i]+6);
      } else if (strncmp(argv[i], "-args=", 6) == 0) {
         if (strcmp(argv[i]+6, "help") == 0) {
            args_help();
            return 0;
         }
         FindMkisofs::SetMkisofsArguments(argv[i]+6);
      } else if (strncmp(argv[i], "-addarg=", 8) == 0) {
         if (strcmp(argv[i]+8, "help") == 0) {
            addarg_help();
            return 0;
         }
         std::string newargs(FindMkisofs::GetPlainMkisofsArguments());
         newargs.append(1, ' ');
         newargs.append(argv[i]+8);
         FindMkisofs::SetMkisofsArguments(newargs);
      } else if (strncmp(argv[i], "-infile=", 8) == 0) {
         if (strcmp(argv[i]+8, "help") == 0) {
            infile_help();
            return 0;
         }
         infile = argv[i]+8;
      } else if (strncmp(argv[i], "-outfile=", 9) == 0) {
         if (strcmp(argv[i]+9, "help") == 0) {
            outfile_help();
            return 0;
         }
         outfile = argv[i]+9;
      } else if (strncmp(argv[i], "-unpacked=", 10) == 0) {
         if (strcmp(argv[i]+10, "help") == 0) {
            unpacked_help();
            return 0;
         }
         unpacked = argv[i]+10;
      } else if (strcmp(argv[i], "-help") == 0) {
         usage_help();
         return 0;
      } else {
         std::cerr << "Unrecognized option: " << argv[i]
                   << " (use -help option for more info)"
                   << std::endl;
         return 2;
      }
   }
   if (! FindMkisofs::ValidMkisofsProg()) {
      std::cerr << FindMkisofs::GetMkisofsProgram()
                << " is not a valid executable, searching for mkisofs."
                << std::endl;
      if (! FindMkisofs::SearchPathForMkisofs()) {
         std::cerr << "Unable to find any mkisofs in PATH" << std::endl;
         return 2;
      }
   }
   if (read_packing_list_from(infile)) {

      // Setup SIGINT to interrupt the packing algorithms as if
      // time had expired.

      struct sigaction act;
      memset((void *)&act, 0, sizeof(act));
      act.sa_handler = Packable::SignalHandler;
      sigaction(SIGINT, &act, NULL);

      if (algos.size() == 0) {
         // If no algorithms given, try all the ones that don't take
         // a long time to run.
         algos.push_back("simple");
         algos.push_back("ffd");
         algos.push_back("luck1000");
      }
      bool found_packing = false;
      for (int i = 0; i < algos.size(); ++i) {
         const char * algorithm = algos[i].c_str();
         if (strcmp(algorithm, "simple") == 0) {
            found_packing |= Packable::SimpleHeuristicPacking();
         } else if (strcmp(algorithm, "brute") == 0) {
            found_packing |= Packable::BruteForceFindPacking();
         } else if (strcmp(algorithm, "ffd") == 0) {
            found_packing |= Packable::FirstFitDecreasing();
         } else if (strncmp(algorithm, "luck", 4) == 0) {
            unsigned long iter = strtoul(algorithm + 4, NULL, 10);
            found_packing |= Packable::MonteCarloPacking(iter);
         } else {
            std::cerr << "Unrecognized packing algorithm: " << algorithm
                      << std::endl;
            return 2;
         }
      }
      if (found_packing) {
         bool rval = print_packing_to(outfile);
         if (rval && (unpacked.size() > 0)) {
            std::ofstream ofs(unpacked.c_str(),
                              std::ios_base::in|std::ios_base::trunc);
            if (ofs.is_open()) {
               Packable::PrintUnpacked(ofs);
            } else {
               std::cerr << "Unable to write " << unpacked << std::endl;
               rval = false;
            }
         }
         if (rval) {
            return Packable::WastesTooMuch() ? 2 : 0;
         }
      }
   }
   return 2;
}

static bool
read_packing_list_from(std::string infile) {
   bool rval = false;
   if (infile.size() == 0) {
      rval = Packable::ReadPackingList(std::cin);
   } else {
      std::ifstream ifs(infile.c_str(), std::ios_base::in);
      if (ifs.is_open()) {
         rval = Packable::ReadPackingList(ifs);
         ifs.close();
      } else {
         std::cerr << "Unable to read " << infile << std::endl;
      }
   }
   return rval;
}

static bool
print_packing_to(std::string outfile) {
   bool rval = true;
   if (outfile.size() == 0) {
      Packable::PrintPacking(std::cout);
   } else {
      std::ofstream ofs(outfile.c_str(), std::ios_base::in|std::ios_base::trunc);
      if (ofs.is_open()) {
         Packable::PrintPacking(ofs);
      } else {
         std::cerr << "Unable to write " << outfile << std::endl;
         rval = false;
      }
   }
   return rval;
}

static void
print_usage_lines(const char ** lines) {
   while (*lines != NULL) {
      std::cerr << *lines++ << std::endl;
   }
}

static const char * algorithm_help_lines[] = {
"-algorithm=name - name the algorithm to try to use to find the packing.",
"This option may be specified multiple times to try multiple algorithms in a",
"single isopack run. If no -algorithm argument is given the default is to use",
"these values:",
"",
"-algorithm=simple -algorithm=ffd -algorithm=luck1000",
"",
"The available algorithms are:",
"",
"simple - a simple algorithm devised with no real underlying theory, but it",
"runs fast and produces a moderately OK packing.",
"",
"brute - a brute force iterate through all permutations packing.  Unless you",
"have a very small set of groups to pack, this will probably never",
"finish. The use of the -time option with this is highly recommended (and it",
"usually finds some good packings in the first few minutes).",
"",
"ffd - first fit descending. Starts with the files sorted biggest first and",
"tries adding them one at a time, skipping to the next smaller file when a",
"file it attempts to add won't fit.",
"",
"luckN - the random Monte Carlo packing algorithm. Picks permutations at",
"random and checks the fit. The 'N' is not literally part of the name, but is",
"replaced with any integer value to indicate how many random packings to",
"try. If you specify -luck0 the algorithm will never terminate, so you need",
"to specify a -time option when you say -luck0.",
NULL
};

static void
algorithm_help() {
   print_usage_lines(algorithm_help_lines);
}

static const char * time_help_lines[] = {
"-time=seconds - specify the max number of seconds any one algorithm is",
"allowed to run (if multiple -algorithm options are given, each separate",
"algorithm is allowed to run this long).",
"",
"append an 'm' or 'M' to the number to specify the time in minutes.",
"",
"append an 'h' or 'H' to the number to specify the time in hours.",
NULL
};

static void
time_help() {
   print_usage_lines(time_help_lines);
}

static const char * close_help_lines[] = {
"-close=sectors - specify number of sectors you are willing to waste (a",
"packing that is \"close enough\"). If a packing is found that is within this",
"many sectors of the media size, the search for the best packing will be",
"terminated immediately (useful together with -time option to avoid taking",
"the full time once you are close enough).",
NULL
};

static void
close_help() {
   print_usage_lines(close_help_lines);
}

static const char * waste_help_lines[] = {
"-waste=sectors - specify max number of sectors you are willing to waste on",
"the media. If the best packing found wastes more space than this, then",
"isopack exits with an error status. (The generated packing is still written",
"to the output).",
NULL
};

static void
waste_help() {
   print_usage_lines(waste_help_lines);
}

static const char * watchdog_help_lines[] = {
"-watchdog=seconds Takes the same arguments as -time= option, but instead of",
"timing the entire run of an algorithm, acts like a \"watchdog timer\",",
"starting over after each packing is found and stopping the algorithm only",
"after it runs this long without finding a better packing.",
NULL
};

static void
watchdog_help() {
   print_usage_lines(watchdog_help_lines);
}

static const char * verbose_help_lines[] = {
"-verbose=int where int is one of:",
"0 - quiet, no extra informat printed to stderr",
"1 - as new packing candidates are found, prints the description (default)",
"2 - 1 + additional information about the progress of the algorithms",
NULL
};

static void
verbose_help() {
   print_usage_lines(verbose_help_lines);
}

static const char * sectors_help_lines[] = {
"-sectors=size where size is the number of 2048 byte sectors, or size is one",
"of the predefined sector sizes for different kinds of media:",
"",
NULL
};

static void
sectors_help() {
   print_usage_lines(sectors_help_lines);
   for (int i = 0; i < NUM_ISO; ++i) {
      std::cerr << std::setw(12) << sector_list[i].name
                << std::setw(10) << sector_list[i].max_sectors
                << "  " << sector_list[i].desc << std::endl;
   }
   std::cerr << std::endl
             << "The default media size if not specified is BD-R.SL"
             << std::endl;
}

static const char * prog_help_lines[] = {
"-prog=file should give the full path name of the mkisofs program",
"to be used by iospack to determine exact size of ISO image files.",
NULL
};

static void
prog_help() {
   print_usage_lines(prog_help_lines);
}

static const char * args_help_lines[] = {
"-args=string Give the argument string that will be passed to mkisofs.",
"The -print-size and -graft-points args will always be prepended to this",
"string, and the -path-list arg (along with temp file holding the list)",
"will always be appended to this string. The arguments you provide",
"here go in the middle.",
"",
"The current argument string is:",
"",
NULL
};

static void
args_help() {
   print_usage_lines(args_help_lines);
   std::cerr << FindMkisofs::GetPlainMkisofsArguments() << std::endl;
}

static const char * addarg_help_lines[] = {
"-addarg=string Give one argument to add to the current mkisofs argument",
"list. This may be called multiple times to add multiple arguments.",
NULL
};

static void
addarg_help() {
   print_usage_lines(addarg_help_lines);
}

static const char * infile_help_lines[] = {
"-infile=file The input file groupings are normally read from stdin, but",
"you can use this option to name a file to read from instead of stdin.",
NULL
};

static void
infile_help() {
   print_usage_lines(infile_help_lines);
}

static const char * outfile_help_lines[] = {
"-outfile=file The path list of the packed files is normally written to",
"stdout, but if you give this option, you can name a file where the packing",
"will be written instead of stdout.",
NULL
};

static void
outfile_help() {
   print_usage_lines(outfile_help_lines);
}

static const char * unpacked_help_lines[] = {
"-unpacked=file Name a file where the path list of files not included in the",
"packing will be written. This output retains the grouping information from",
"the input, so it can be fed into isopack again to determine a new packing",
"for the remaining files (useful if you want to backup several discs worth",
"of files).",
NULL
};

static void
unpacked_help() {
   print_usage_lines(unpacked_help_lines);
}

static const char * usage_help_lines[] = {
"Usage: isopack -algorithm=name... -time=seconds -verbose=int -sectors=size",
"",
"-algorithm= Give the name of a packing algorithm (option may be repeated)",
"-time=      Give max seconds allowed for algorithm to run",
"-close=     How many sectors are close enough to stop searching",
"-waste=     How many sectors are you willing to waste",
"-watchdog=  Like -time=, but starts over each time a new packing is found",
"-verbose=   Verbosity level: 0 is quiet, 2 is wordy",
"-sectors=   Give max 2048 byte sectors in ISO or predefined name",
"-prog=      Name the mkisofs program for isopack to invoke",
"-args=      Argument string to use when calling mkisofs",
"-addarg=    Give extra arguments to add to the mkisofs argument list",
"-infile=    Give file to read from rather than the default stdin",
"-outfile=   Give file to write to rather than the default stdout",
"-unpacked=  Files from input that were not packed may be recorded here",
"",
"isopack reads a path list on stdin, removes enough files to fit in",
"the desired ISO image size, and writes new list on stdout.",
"",
"The word \"help\" may be given as the value of any argument for more info.",
NULL
};

static void
usage_help() {
   print_usage_lines(usage_help_lines);
}
