
// Written Dec 2011 Tom Horsley
// Released to the universe with no license restrictions of any kind.

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <algorithm>
#include <fstream>
#include <iomanip>
#include "packable.h"
#include "find-mkisofs.h"

int                    Packable::verbose_level = 1;
unsigned long          Packable::max_seconds = 0;
bool                   Packable::watchdog_timer = false;
time_t                 Packable::cutoff_time;
bool                   Packable::out_of_time = false;
unsigned long long     Packable::max_sectors = 12219392ull;
unsigned long long     Packable::max_bytes = 12219392ull * 2048ull;
unsigned long long     Packable::close_enough_bytes = 0;
unsigned long long     Packable::max_waste_bytes = (unsigned long long)(-1ll);

std::vector<Packable*> Packable::packing_list;

unsigned long long     Packable::best_byte_size = 0;
unsigned long long     Packable::best_iso_size = 0;
int                    Packable::best_file_count = 0;
std::vector<Packable*> Packable::best_packing_list;
std::string            Packable::best_algorithm;
time_t                 Packable::best_seconds;

time_t                 Packable::start_time;
std::string            Packable::cur_algorithm;

std::string Packable::tempfile;

int Packable::next_seqnum = 0;

void
Packable::SetMaxSectors(unsigned long long init_max_sectors) {
   max_sectors = init_max_sectors;
   max_bytes = max_sectors * 2048ull;
}

void
Packable::SetSecondsLimit(unsigned long init_max_seconds) {
   max_seconds = init_max_seconds;
}

void
Packable::SetWatchdogSecondsLimit(unsigned long init_max_seconds) {
   SetSecondsLimit(init_max_seconds);
   watchdog_timer = true;
}

void
Packable::SetCloseEnough(unsigned long long close_enough) {
   close_enough_bytes = close_enough * 2048ull;
}

void
Packable::SetMaxWaste(unsigned long long max_waste) {
   max_waste_bytes = max_waste * 2048ull;
}

bool
Packable::WastesTooMuch() {
   return (max_bytes - best_iso_size) > max_waste_bytes;
}

void
Packable::SetVerbose(int init_verbose_level) {
   verbose_level = init_verbose_level;
}

bool
Packable::ReadPackingList(std::istream & input) {
   int state = 0;
   std::string line;
   std::string start_normal_group("[");
   std::string start_mustadd_group(start_normal_group);
   start_mustadd_group.append("!");
   std::string end_group("]");
   std::vector<std::string> current_group;
   bool must_pack_group;

   while (input.good()) {
      getline(input, line);
      if (line.size() > 0) {
         switch(state) {
         case 0: // just accumulating lines not in any group
            if (line == start_normal_group) {
               current_group.clear();
               must_pack_group = false;
               state = 1;
            } else if (line == start_mustadd_group) {
               current_group.clear();
               must_pack_group = true;
               state = 1;
            } else if (line == end_group) {
               std::cerr
                  << "Packing list: end group line with no matching begin."
                  << std::endl;
               return false;
            } else {
               current_group.push_back(line);
               if (! AddPackable(current_group, false)) {
                  return false;
               }
               current_group.clear();
            }
            break;
         case 1: // Accumlating group
            if ((line == start_normal_group) ||
                (line == start_mustadd_group)) {
               std::cerr << "Packing list: nested groups not allowed."
                         << std::endl;
               return false;
            } else if (line == end_group) {
               if (current_group.size() == 0) {
                  std::cerr << "Packing list: empty group makes no sense."
                            << std::endl;
                  return false;
               } else {
                  if (! AddPackable(current_group, must_pack_group)) {
                     return false;
                  }
                  current_group.clear();
                  must_pack_group = false;
                  state = 0;
               }
            } else {
               current_group.push_back(line);
            }
            break;
         }
      }
   }
   if (packing_list.size() == 0) {
      std::cerr << "No files to pack." << std::endl;
      return false;
   }
   return true;
}

bool
Packable::BruteForceFindPacking() {

   if ((close_enough_bytes != 0) &&
       (close_enough_bytes >= (max_bytes - best_iso_size))) {

      // Previous algorithm apparently got close enough. No need to
      // run this at all.

      return true;
   }

   cur_algorithm = "brute";
   start_time = time(NULL);

   if (verbose_level > 1)
   std::cerr << "Trying brute force packing." << std::endl;

   SetCutoffTime();

   // Sort the list of packable items so that the MustPack() items
   // are first, then after that, they are sorted with the largest
   // byte size first.

   std::sort(packing_list.begin(), packing_list.end(), PackableLess);

   // Find the total raw byte size of all the files.

   unsigned long long total_size = 0;
   for (int i = 0; i < packing_list.size(); ++i) {
      total_size += packing_list[i]->ByteSize();
   }

   // Now find out how many of the biggest files I have to remove to have
   // the slightest chance of fitting the files in the max size iso
   // image.

   int remove_big_count = 0;
   unsigned long long current_size = total_size;
   MarkAllPacked();
   int i = 0;
   for ( ; i < packing_list.size(); ++i) {
      if (current_size <= max_bytes) {
         break;
      }
      if (! packing_list[i]->MustPack()) {
         packing_list[i]->MarkPacked(false);
         current_size -= packing_list[i]->ByteSize();
         remove_big_count++;
      }
   }

   // That is a good estimate of the number of files to remove, but now
   // get the true size by trying the packing and finding out how
   // big the actual ISO image would be. Continue to remove files
   // till the actual ISO size is within range.

   for ( ; ; ) {
      current_size = FindISOSizeOfPacking();
      if (current_size <= max_bytes) {
         CheckNewBestSize(current_size);
         break;
      }
      ++i;
      if (i >= packing_list.size()) {
         break;
      }
      packing_list[i]->MarkPacked(false);
      remove_big_count++;
   }

   if (current_size > max_bytes) {
      std::cerr << "Unable to remove enough groups to fit within max ISO size."
                << std::endl;
      return false;
   }

   if (verbose_level > 1)
   std::cerr << "I need to remove at least " << remove_big_count
             << " groups of " << packing_list.size()
             << " to fit in max ISO size." << std::endl;

   // Then find out how many of the smallest files I have to remove to
   // have a possible chance to fit the files in the iso image.

   int remove_small_count = 0;
   current_size = total_size;
   i = packing_list.size() - 1;
   for ( ; i >= 0; --i) {
      if (current_size <= max_bytes) {
         break;
      }
      if (packing_list[i]->MustPack()) {
         break;
      }
      current_size -= packing_list[i]->ByteSize();
      remove_small_count++;
   }

   // And continue removing smallest files till the actual ISO size
   // estimate fits in max size.

   for ( ; ; ) {
      current_size = FindISOSizeOfPacking();
      if (current_size <= max_bytes) {
         CheckNewBestSize(current_size);
         break;
      }
      --i;
      if (i < 0) {
         break;
      }
      if (packing_list[i]->MustPack()) {
         break;
      }
      packing_list[i]->MarkPacked(false);
      remove_small_count++;
   }

   if (verbose_level > 1)
   std::cerr << "I need to remove at most " << remove_small_count
             << " groups to fit in max ISO size." << std::endl;

   // Now try removing different combinations of files to find the
   // best set of files we can remove to leave the remaining files
   // filling up the ISO image with the least wasted space.

   int first_index = 0;
   while (packing_list[first_index]->MustPack()) {
      ++first_index;
   }
   for (int remove_count = remove_small_count;
        remove_count >= remove_big_count;
        --remove_count) {
      if (verbose_level > 1)
      std::cerr << "Try removing " << remove_count << " groups" << std::endl;
      MarkAllPacked();
      int last_index = packing_list.size() - 1;
      int temp = remove_count;
      while (temp-- > 0) {
         packing_list[last_index--]->MarkPacked(false);
      }

      // Try moving the top not_packed file through all the
      // positions of the packing_list starting at last_index+1
      // (where it currently is) and moving up to first_index
      // (which is the last file we can try unpacking).
      // Recursively do the same for the other unpacked files
      // below the top one.

      TryEachFile(first_index, last_index+1, remove_count - 1);
      if (out_of_time) {
         break;
      }
   }

   TrySmallFiles();

   return best_packing_list.size() > 0;
}

bool
Packable::SimpleHeuristicPacking() {

   if ((close_enough_bytes != 0) &&
       (close_enough_bytes >= (max_bytes - best_iso_size))) {

      // Previous algorithm apparently got close enough. No need to
      // run this at all.

      return true;
   }

   cur_algorithm = "simple";
   start_time = time(NULL);

   if (verbose_level > 1)
   std::cerr << "Trying simple heuristic packing." << std::endl;

   SetCutoffTime();

   // Sort the list of packable items so that the MustPack() items
   // are first, then after that, they are sorted with the largest
   // byte size first.

   std::sort(packing_list.begin(), packing_list.end(), PackableLess);
   MarkAllPacked();
   unsigned long long iso_size = FindSomeSizeOfPacking();
   if (iso_size <= max_bytes) {
      if (verbose_level > 1)
      std::cerr << "All files fit within max ISO image size." << std::endl;
      CheckNewBestSize(iso_size);
      return true;
   }
   int item_count = packing_list.size();
   for (int n = 1; n < (item_count-1); ++n) {
      if (TryWithNFewerItems(n)) {
         break;
      }
      if (out_of_time) {
         break;
      }
   }

   TrySmallFiles();

   return best_packing_list.size() > 0;
}

bool
Packable::FirstFitDecreasing() {

   if ((close_enough_bytes != 0) &&
       (close_enough_bytes >= (max_bytes - best_iso_size))) {

      // Previous algorithm apparently got close enough. No need to
      // run this at all.

      return true;
   }

   cur_algorithm = "ffd";
   start_time = time(NULL);

   if (verbose_level > 1)
   std::cerr << "Trying first fit decreasing packing." << std::endl;

   SetCutoffTime();

   // OK, sort with MustPack items first and in decreasing size after
   // that.

   std::sort(packing_list.begin(), packing_list.end(), PackableLess);

   int first_big_file = 0;
   while (packing_list[first_big_file]->MustPack()) {
      ++first_big_file;
   }
   while (first_big_file < packing_list.size()) {
      if (verbose_level > 1)
      std::cerr << "Trying with first big item: " << first_big_file
                << std::endl;
      MarkAllPacked(false);
      for (int i = first_big_file; i < packing_list.size(); ++i) {
         packing_list[i]->MarkPacked(true);
         unsigned long long iso_size = FindSomeSizeOfPacking();
         if (iso_size < max_bytes) {
            CheckNewBestSize(iso_size);
         } else {

            // This one don't fit, throw it back out.

            packing_list[i]->MarkPacked(false);
         }
         if (out_of_time) {
            break;
         }
      }
      ++first_big_file;

      // If all the remaining items would fit, no point in continuing
      // to increment first_big_file.

      MarkAllPacked(false);
      for (int i = first_big_file; i < packing_list.size(); ++i) {
         packing_list[i]->MarkPacked(true);
      }
      unsigned long long iso_size = FindSomeSizeOfPacking();
      if (iso_size < max_bytes) {
         CheckNewBestSize(iso_size);
         break;
      }
      if (out_of_time) {
         break;
      }
   }

   TrySmallFiles();

   return best_packing_list.size() > 0;
}

bool
Packable::MonteCarloPacking(unsigned long iter) {

   if ((close_enough_bytes != 0) &&
       (close_enough_bytes >= (max_bytes - best_iso_size))) {

      // Previous algorithm apparently got close enough. No need to
      // run this at all.

      return true;
   }

   cur_algorithm = "luck";
   start_time = time(NULL);

   if (verbose_level > 1)
   std::cerr << "Trying Monte Carlo (random) packing." << std::endl;

   if ((iter == 0) && (max_seconds == 0)) {
      if (verbose_level > 1)
      std::cerr << "Iteration count set to 1 since no time limit was given."
                << std::endl;
      iter = 1;
   }

   SetCutoffTime();

   // Find a nice random seed for drand48 to start with

   long seedval;
   bool got_seed = false;
   int fd = open("/dev/urandom", O_RDONLY);
   if (fd != -1) {
      if (read(fd, (void *)&seedval, sizeof(seedval)) == sizeof(seedval)) {
         got_seed = true;
      }
      close(fd);
   }
   if (! got_seed) {

      // Apparently /dev/urandom didn't work, let's just grab
      // time of day instead and mush it up with pid and a stack
      // address (which may be random if kernel randomizes stacks).

      seedval = (long int)time(NULL);
      seedval ^= (((long int)getpid()) <<
                  ((((unsigned long)((void *)&seedval)) >> 17) & 0x7));
   }
   srand48(seedval);

   // Now try finding random permutations that fit

   unsigned long cur_count = 0;
   while ((iter == 0) || (cur_count++ < iter)) {
      std::sort(packing_list.begin(), packing_list.end(), PackableLess);
      MarkAllPacked(false);
      int try_next = 0;
      while (packing_list[try_next]->MustPack()) {
         ++try_next;
      }
      int num_left;
      while ((num_left = (packing_list.size() - try_next)) > 0) {
         int rndx = try_next + (int)((drand48() * (double)num_left));
         packing_list[rndx]->MarkPacked(true);
         unsigned long long iso_size = FindSomeSizeOfPacking();
         if (iso_size <= max_bytes) {
            CheckNewBestSize(iso_size);

            // The item at rndx was successfully packed, move it up
            // to the try_next position and increment try_next to point
            // to new remaining set of items not yet packed.

            Packable * p = packing_list[rndx];
            for (int i = rndx - 1; i >= try_next; --i) {
               packing_list[i+1] = packing_list[i];
            }
            packing_list[try_next] = p;
            ++try_next;
         } else {

            // This randomly selected file didn't fit. Since everything
            // at lower index values is either already packed or is a bigger
            // file, we can just set try_next to the item after rndx.

            try_next = rndx + 1;
         }
         if (out_of_time) break;
      }
      if (out_of_time) break;
   }

   TrySmallFiles();

   return best_packing_list.size() > 0;
}

void
Packable::PrintPacking(std::ostream & output) {

   if (verbose_level > 0) {
      unsigned long long best_iso_sectors = best_iso_size / 2048ull;
      unsigned long long unused_sectors = max_sectors - best_iso_sectors;
      double media_percent = ((double)best_iso_sectors)/((double)max_sectors);
      std::cerr << "Best packing found contains " << best_byte_size
                << " bytes of data in " << best_file_count
                << " files." << std::endl
                << "Unused sectors on media: " << unused_sectors
                << " of " << max_sectors << " ("
                << std::setprecision (3) << std::fixed << media_percent*100.0
                << "% full)" << std::endl
                << "Packing found in " << best_seconds << " seconds by "
                << best_algorithm << " algorithm." << std::endl;
   }

   // Sort the output list back into the same relative sequence as the
   // input list.

   std::sort(best_packing_list.begin(), best_packing_list.end(),
             Packable::SeqnumLess);

   // Print the lines of the best packing list we found

   for (int i = 0; i < best_packing_list.size(); ++i) {
      Packable * p = best_packing_list[i];
      for (int j = 0; j < p->lines.size(); ++j) {
         output << p->lines[j] << std::endl;
      }
   }
}

void
Packable::PrintUnpacked(std::ostream & output) {

   // Sort the list back into original sequence.

   std::sort(packing_list.begin(), packing_list.end(),
             Packable::SeqnumLess);

   // Print the lines and the grouping syntax for groups we didn't include
   // in the best packing. (You can feed this into isopack again to fill up
   // the next ISO image).

   MarkAllPacked(false);
   for (int i = 0; i < best_packing_list.size(); ++i) {
      best_packing_list[i]->MarkPacked(true);
   }

   for (int i = 0; i < packing_list.size(); ++i) {
      Packable * p = packing_list[i];
      if (! p->IsPacked()) {
         if (p->lines.size() > 1) {
            output << "[" << std::endl;
         }
         for (int j = 0; j < p->lines.size(); ++j) {
            output << p->lines[j] << std::endl;
         }
         if (p->lines.size() > 1) {
            output << "]" << std::endl;
         }
      }
   }
}

void
Packable::SignalHandler(int num) {
   // Don't do anything foolish in interrupt handler - just set the flag.
   out_of_time = true;
}

Packable::Packable(
   const std::vector<std::string> & init_lines,
   unsigned long long init_byte_size,
   bool init_must_pack)
 : lines(init_lines),
   byte_size(init_byte_size),
   is_packed(init_must_pack),
   must_pack(init_must_pack),
   seqnum(next_seqnum++) {
}

Packable::~Packable() {
   return;
}

void
Packable::MarkPacked(bool init_is_packed) {
   if (! must_pack) is_packed = init_is_packed;
}

void
Packable::TryEachFile(int low_index, int high_index, int remaining_items) {
   if (remaining_items < 0) {
      return;
   }
   if (remaining_items == 0) {

      // OK, I'm the last bit to toggle. I get to actually try the
      // packings to see which ones are the best for each position I
      // can fit in.

      for ( ; ; ) {
         unsigned long long test_size = FindSomeSizeOfPacking();
         if (test_size <= max_bytes) {
            CheckNewBestSize(test_size);
         }
         if (low_index == high_index) {
            // I can't move my mark up any, get out of loop.
            break;
         }
         packing_list[high_index]->MarkPacked(true);
         --high_index;
         packing_list[high_index]->MarkPacked(false);
         if (out_of_time) {
            break;
         }
      }
   } else {

      // Try moving my mark bit through all my available positions
      // and do recursive call to get lower level mark bits to
      // try all their positions below my current position.

      int cur_index = high_index;
      for ( ; ; ) {
         TryEachFile(cur_index+1, high_index+1, remaining_items-1);
         if (cur_index == low_index) {
            break;
         }
         packing_list[cur_index]->MarkPacked(true);
         --cur_index;
         packing_list[cur_index]->MarkPacked(false);
         for (int i = cur_index+1; i < packing_list.size(); ++i) {
            packing_list[i]->MarkPacked(i < high_index+1);
         }
         if (out_of_time) {
            break;
         }
      }
   }
}

bool
Packable::PackableLess(const Packable * a, const Packable * b) {
   if (a->MustPack() != b->MustPack()) {
      return a->MustPack();
   }
   return a->ByteSize() >= b->ByteSize();
}

bool
Packable::SeqnumLess(const Packable * a, const Packable * b) {
   return a->seqnum < b->seqnum;
}

bool
Packable::AddPackable(
   const std::vector<std::string> & init_lines,
   bool init_must_pack) {
   unsigned long long byte_size = 0;
   for (int i = 0; i < init_lines.size(); ++i) {
      std::string file = GetFileName(init_lines[i]);
      struct stat statbuf;
      if (::stat(file.c_str(), &statbuf) != 0) {
         std::cerr << "Packing list: cannot stat " << file << std::endl;
         return false;
      }
      if (! S_ISREG(statbuf.st_mode)) {
         std::cerr << "Packing list: " << file << " is not a regular file."
                   << std::endl;
         return false;
      }
      byte_size += statbuf.st_size;
   }
   packing_list.push_back(new Packable(init_lines, byte_size, init_must_pack));
   return true;
}

std::string
Packable::GetFileName(const std::string & line) {
   std::string rval;
   const char * s = line.c_str();
   bool seen_equal = false;
   bool seen_escape = false;
   char c;
   while ((c = *s++) != '\0') {
      if ((! seen_escape) && (! seen_equal) && (c == '=')) {
         seen_equal = true;
      } else if ((! seen_escape) && (c == '\\')) {
         seen_escape = true;
      } else {
         if (seen_equal) {
            rval.append(1, c);
         }
         seen_escape = false;
      }
   }
   return rval;
}

void
Packable::MarkAllPacked(bool init_is_packed) {
   for (int i = 0; i < packing_list.size(); ++i) {
      packing_list[i]->MarkPacked(init_is_packed);
   }
}

unsigned long long
Packable::FindISOSizeOfPacking() {
   if (tempfile.size() == 0) {

      // This is first time, create temp file name with mkstemp,
      // and record the temp name for reuse on subsequent calls.

      char tempname[50];
      ::strcpy(tempname, "/var/tmp/isopackXXXXXX");
      int fd = ::mkstemp(tempname);
      if (fd == -1) {
         std::cerr << "ACK! I can't create a temp file!" << std::endl;
         ::exit(2);
      } else {
         tempfile.append(tempname);
         close(fd);
      }
   }
   std::ofstream f;
   f.open(tempfile.c_str(), std::ios_base::trunc|std::ios_base::out);
   if (! f.is_open()) {
      std::cerr << "ACK! I can't write on temp file " << tempfile << std::endl;
      ::exit(2);
   }
   for (int i = 0; i < packing_list.size(); ++i) {
      Packable * p = packing_list[i];
      if (p->IsPacked()) {
         for (int j = 0; j < p->lines.size(); ++j) {
            f << p->lines[j] << std::endl;
         }
      }
   }
   f.close();

   std::string cmd(FindMkisofs::GetMkisofsProgram());
   cmd.append(" ");
   cmd.append(FindMkisofs::GetMkisofsArguments());
   cmd.append(" ");
   cmd.append(tempfile);
   cmd.append(" 2>/dev/null");
   FILE * r = ::popen(cmd.c_str(), "r");
   unsigned long long rval = 0;
   if (r != NULL) {
      char buf[50];
      while (::fgets(buf, sizeof(buf), r) != NULL) {
         char * endbuf;
         unsigned long long temp = 0;
         temp = ::strtoull(buf, &endbuf, 10);
         if ((*endbuf == '\n') || (*endbuf == '\0')) {
            rval = temp;
         }
      }
      if ((::pclose(r) != 0) || (rval == 0)) {
         std::cerr << "mkisofs does not appear to run correctly:" << std::endl;
         std::string cmd(FindMkisofs::GetMkisofsProgram());
         cmd.append(" ");
         cmd.append(FindMkisofs::GetMkisofsArguments());
         cmd.append(" ");
         cmd.append(tempfile);
         system(cmd.c_str());
         exit(2);
      }
   }
   rval *= (unsigned long long)2048;
   return rval;
}

unsigned long long
Packable::FindFileSizeOfPacking() {

   // Lower overhead than FindISOSizeOfPacking() and no reason to
   // invoke FindISOSizeOfPacking() if this is already too big.

   unsigned long long rval = 0;
   for (int i = 0; i < packing_list.size(); ++i) {
      if (packing_list[i]->IsPacked()) {
         rval += packing_list[i]->ByteSize();
      }
   }
   return rval;
}

unsigned long long
Packable::FindSomeSizeOfPacking() {
   unsigned long long rval = FindFileSizeOfPacking();
   if (rval <= max_bytes) {
      // Close enough that we may need a second opinion, but not if
      // rval isn't as big as the best packing we already found.
      if (rval >= best_byte_size) {
         rval = FindISOSizeOfPacking();
      }
   }
   if ((! out_of_time) && (cutoff_time != 0)) {
      out_of_time = (time(NULL) >= cutoff_time);
      if (out_of_time) {
         if (verbose_level > 1)
         std::cerr << "Time limit expired." << std::endl;
      }
   }
   return rval;
}

void
Packable::CheckNewBestSize(unsigned long long iso_size) {
   unsigned long long file_sizes = 0;
   int file_count = 0;
   for (int i = 0; i < packing_list.size(); ++i) {
      if (packing_list[i]->IsPacked()) {
         file_sizes += packing_list[i]->ByteSize();
         file_count += packing_list[i]->lines.size();
      }
   }
   if (file_sizes > best_byte_size) {

      // Got new best packing (which we define by the actual number of bytes
      // in all the files in the collection).

      best_algorithm = cur_algorithm;
      best_seconds = time(NULL) - start_time;
      best_byte_size = file_sizes;
      best_iso_size = iso_size;
      best_file_count = file_count;
      best_packing_list.clear();
      for (int i = 0; i < packing_list.size(); ++i) {
         if (packing_list[i]->IsPacked()) {
            best_packing_list.push_back(packing_list[i]);
         }
      }
      if (verbose_level > 0)
      std::cerr << "Found new best packing with " << best_packing_list.size() <<
                   " groups, " << best_file_count << " files at " <<
                   best_seconds << " seconds," <<
                   std::endl << "Total bytes in files " << best_byte_size <<
                   std::endl << "Total bytes in ISO image " << best_iso_size <<
                   std::endl;
      if (watchdog_timer) {
         SetCutoffTime();
      }
      if ((close_enough_bytes != 0) &&
          (close_enough_bytes >= (max_bytes - best_iso_size))) {
         if (verbose_level > 0)
         std::cerr << "Stopping search because close enough match found."
                   << std::endl;
         out_of_time = true; // use same flag as time to stop the loops
      }
      if (best_packing_list.size() == packing_list.size()) {
         std::cerr << "All files fit in media." << std::endl;
         out_of_time = true;
      }
   }
}

bool
Packable::TryWithNFewerItems(int n) {
   if (verbose_level > 1)
   std::cerr << "Try packing with " << n << " fewer groups." << std::endl;
   int num_files = packing_list.size();
   int i;
   for (i = 0; i < (num_files - n); ++i) {
      packing_list[i]->MarkPacked(true);
   }
   for ( ; i < num_files; ++i) {
      packing_list[i]->MarkPacked(false);
   }
   unsigned long long iso_size = FindSomeSizeOfPacking();
   if (iso_size <= max_bytes) {
      CheckNewBestSize(iso_size);
      return true;
   }
   while (ToggleMarkedFiles()) {
      iso_size = FindSomeSizeOfPacking();
      if (iso_size <= max_bytes) {
         CheckNewBestSize(iso_size);
      }
   }
   return false;
}

bool
Packable::ToggleMarkedFiles() {
   int num_files = packing_list.size();
   int i;

   // skip over all leading files that already have on_disk false

   for (i = 0; i < num_files; ++i) {
      if (packing_list[i]->IsPacked()) {
         break;
      }
   }

   // i now points to first file where on_disk is true, find file
   // after that (if any) where on_disk is false and move the false
   // flag up one. (if there is no such file, we are done and should
   // return false).

   bool moved = false;
   for (++i ; i < num_files; ++i) {
      if (! packing_list[i]->IsPacked()) {
         moved = true;
         packing_list[i-1]->MarkPacked(false);
         packing_list[i]->MarkPacked(true);
         break;
      }
   }
   return moved;
}

void
Packable::SetCutoffTime() {
   out_of_time = false;
   if (max_seconds != 0) {
      cutoff_time = time(NULL) + max_seconds;
   }
}

void
Packable::TryNearbyFiles() {
   // Try swapping a currently packed item with the next larger
   // unpacked item to see if a (possibly) small increment in the
   // size can still fit.

   if (verbose_level > 1)
   std::cerr << "Check to see if I can move up to larger groups." << std::endl;

   std::sort(packing_list.begin(), packing_list.end(), PackableLess);
   bool successful_swap;
   do {
      successful_swap = false;
      for (int i = packing_list.size()-1; i >= 0; --i) {
         if (packing_list[i]->IsPacked()) {
            if (packing_list[i]->MustPack()) {
               // We have run out of optionally packed items.
               break;
            } else {
               for (int j = i-1; j >= 0; --j) {
                  if (! packing_list[j]->IsPacked()) {
                     // OK, we found the next largest unpacked item.
                     // Try the sawp and see if we still fit.
                     packing_list[j]->MarkPacked(true);
                     packing_list[i]->MarkPacked(false);
                     unsigned long long iso_size = FindSomeSizeOfPacking();
                     if (iso_size <= max_bytes) {
                        CheckNewBestSize(iso_size);
                        successful_swap = true;
                     } else {
                        packing_list[j]->MarkPacked(false);
                        packing_list[i]->MarkPacked(true);
                     }
                     break;
                  }
               }
               if (successful_swap) {
                  break;
               }
            }
         }
      }
   } while(successful_swap);
}

void
Packable::TrySmallFiles() {

   // First try checking to see if we can bump up some items to next
   // bigger items without overflowing ISO.

   TryNearbyFiles();

   // As last step to tweak any given packing algorithm, try
   // adding the smallest files not yet added until the size
   // gets too big. (This probably won't ever do anything, but
   // it gives me a common location to cleanup the temp file
   // as well).

   if (verbose_level > 1)
   std::cerr << "Check to see if I can add smallest groups." << std::endl;
   MarkAllPacked(false);
   for (int i = 0; i < best_packing_list.size(); ++i) {
      best_packing_list[i]->MarkPacked(true);
   }
   std::sort(packing_list.begin(), packing_list.end(), PackableLess);
   for (int i = packing_list.size() - 1; i >= 0; --i) {
      if (! packing_list[i]->IsPacked()) {
         packing_list[i]->MarkPacked(true);
         unsigned long long iso_size = FindSomeSizeOfPacking();
         if (iso_size <= max_bytes) {
            CheckNewBestSize(iso_size);
         } else {
            // That's it. The smallest remaining file won't fit,
            // call it a day.
            break;
         }
      }
   }

   // Once we have the packing, we don't need the temp file anymore

   if (tempfile.size() > 0) {
      unlink(tempfile.c_str());
      tempfile.clear();
   }
}
