#ifndef packable_h
#define packable_h

// Written Dec 2011 Tom Horsley
// Released to the universe with no license restrictions of any kind.

#include <string>
#include <vector>
#include <iostream>
#include <time.h>

// Class Packable holds information about a single "packable" item. This
// is a collection of -path-list lines that must always be kept together.
//
// The static functions in this class form the public interface for
// running the packing algorithm.
//
class Packable {
public:

      // Before calling one of the packing routines, call this
      // to set the number of 2048 byte sectors that are the max
      // size of the ISO image you wish to create.
      //
      // The default size is 12219392 (the size of max unformatted
      // BD-R disk).
      //
   static void
   SetMaxSectors(unsigned long long init_max_sectors);

      // Before calling one of the packing routines, you can call
      // this routine to put a limit on the number of seconds the
      // algorithm takes. When time expires, the best packing
      // it has found so far is the one it goes with.
      //
      // The default is 0, meaning no time limit.
      //
   static void
   SetSecondsLimit(unsigned long init_max_seconds);

      // Set max time and enable the watchdog_timer flag to treat
      // max time as time since last new packing found rather than
      // time since algorithm started.
      //
   static void
   SetWatchdogSecondsLimit(unsigned long init_max_seconds);

      // How many sectors is "close enough" for us to stop searching
      // for a better match?
      //
      // Default of 0 means I want best possible match.
      //
   static void
   SetCloseEnough(unsigned long long close_enough);

      // Say how many sectors you are willing to waste on the media.
      // If the best packing found wastes more sectors than this,
      // then WastesTooMuch() will return true.
      //
   static void
   SetMaxWaste(unsigned long long max_waste);

      // Check packing against the max waste value, return true
      // if the packing wastes too much space.
      //
   static bool
   WastesTooMuch();

      // Before calling one of the packing routines, you can set
      // the level of verbosity - 0 means don't say nuttin, higher
      // numbers print interesting tidbits during the discovery of
      // the packing.
      //
      // The default is 1 (which lists the new best packing
      // stats as each new packing is found).
      //
   static void
   SetVerbose(int init_verbose_level);

      // Read a -graft-points style path list from the input stream.
      // Recognize additional syntax to group files together which
      // must always be written together.
      //
      // [
      // pack-list lines...
      // ]
      //
      // defines a group of pack-list lines that will be kept together
      // or all left off.
      //
      // [!
      // pack-list lines...
      // ]
      //
      // defines a group of pack-list lines that must always be
      // written to the ISO (the packing algorithm will always include them).
      //
   static bool
   ReadPackingList(std::istream & input);

      // This will probably take forever, but by golly it will find
      // the best possible packing.
      //
   static bool
   BruteForceFindPacking();

      // This takes very little time, just going up the list of files to
      // find a combination that fits by leaving out different files when
      // it gets too big, but not investigating every possible combination.
      // 
   static bool
   SimpleHeuristicPacking();

      // One of the popular heuristics for bin packing seems to be
      // sorting biggest first, then picking from biggest to smallest
      // till files won't fit, skipping that file and trying next one, etc.
      // This does that, but it also tries again leaving off the biggest
      // file to see if the next set of firt fits works better, etc.
      //
   static bool
   FirstFitDecreasing();

      // Randomly select items from list till I can't pick another
      // without going over the size limit.
      //
      // Repeat for iter iterations or till time runs out if iter
      // is zero.
      //
   static bool
   MonteCarloPacking(unsigned long iter);

      // Print the best packing found in the same -graft-points style
      // as the input but without the [] grouping syntax.
      //
   static void
   PrintPacking(std::ostream & output);

      // Print the files not packed into the best packing. Include
      // the original input grouping syntax so this can be used for
      // another isopack run to add the remaining files.
      //
   static void
   PrintUnpacked(std::ostream & output);

      // Regiser this as an interrupt handler, and if the process
      // is interrupted, it will be treated as though the time
      // expired.
      //
   static void
   SignalHandler(int num);

private:

   Packable(const std::vector<std::string> & init_lines,
            unsigned long long init_byte_size,
            bool init_must_pack = false);

   Packable(const Packable & cpy);

   Packable &
   operator=(const Packable & cpy);

   ~Packable();

   void
   MarkPacked(bool init_is_packed);

   bool
   IsPacked() const {
         return is_packed;
      }

   bool
   MustPack() const {
         return must_pack;
      }

   unsigned long long
   ByteSize() const {
         return byte_size;
      }

   static void
   TryEachFile(int low_index, int high_index, int remaining_items);

   static bool
   PackableLess(const Packable * a, const Packable * b);

   static bool
   SeqnumLess(const Packable * a, const Packable * b);

   static bool
   AddPackable(const std::vector<std::string> & init_lines,
               bool init_must_pack = false);

   static std::string
   GetFileName(const std::string & line);

   static void
   MarkAllPacked(bool init_is_packed = true);

   static unsigned long long
   FindISOSizeOfPacking();

   static unsigned long long
   FindFileSizeOfPacking();

   static unsigned long long
   FindSomeSizeOfPacking();

   static void
   CheckNewBestSize(unsigned long long iso_size);

   static bool
   TryWithNFewerItems(int n);

   static bool
   ToggleMarkedFiles();

   static void
   SetCutoffTime();

   static void
   TryNearbyFiles();

   static void
   TrySmallFiles();

   static int verbose_level;
   static unsigned long max_seconds;
   static bool watchdog_timer;
   static time_t cutoff_time;
   static bool out_of_time;
   static unsigned long long max_sectors;
   static unsigned long long max_bytes;
   static unsigned long long close_enough_bytes;
   static unsigned long long max_waste_bytes;

   static std::vector<Packable*> packing_list;

   static unsigned long long     best_byte_size;
   static unsigned long long     best_iso_size;
   static int                    best_file_count;
   static std::vector<Packable*> best_packing_list;
   static std::string            best_algorithm;
   static time_t                 best_seconds;

   static time_t                 start_time;
   static std::string            cur_algorithm;

   static std::string tempfile;

   static int next_seqnum;

   std::vector<std::string> lines;

   unsigned long long byte_size;

   int seqnum;

   bool is_packed;
   bool must_pack;
};

#endif // packable_h
